@@ -49,9 +49,6 @@
 
 package com.openexchange.filestore.s3.internal;
 
-import static com.openexchange.filestore.utils.PropertyNameBuilder.optBoolProperty;
-import static com.openexchange.filestore.utils.PropertyNameBuilder.optProperty;
-import static com.openexchange.filestore.utils.PropertyNameBuilder.requireProperty;
 import java.io.FileInputStream;
 import java.io.FileNotFoundException;
 import java.io.IOException;
@@ -67,6 +64,8 @@ import java.security.UnrecoverableKeyException;
 import java.security.cert.Certificate;
 import java.security.cert.CertificateException;
 import java.util.Enumeration;
+import java.util.HashMap;
+import java.util.Map;
 import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 import com.amazonaws.AmazonClientException;
@@ -85,17 +84,24 @@ import com.amazonaws.services.s3.model.CreateBucketRequest;
 import com.amazonaws.services.s3.model.CryptoConfiguration;
 import com.amazonaws.services.s3.model.EncryptionMaterials;
 import com.amazonaws.services.s3.model.Region;
+import com.amazonaws.services.s3.model.SetBucketPolicyRequest;
 import com.amazonaws.services.s3.model.StaticEncryptionMaterialsProvider;
+import com.amazonaws.auth.policy.Policy;
+import com.amazonaws.auth.policy.Principal;
+import com.amazonaws.auth.policy.Resource;
+import com.amazonaws.auth.policy.Statement;
+import com.amazonaws.auth.policy.actions.S3Actions;
+import com.amazonaws.auth.policy.conditions.BooleanCondition;
+import com.amazonaws.auth.policy.conditions.StringCondition;
+import com.amazonaws.auth.policy.conditions.StringCondition.StringComparisonType;
 import com.openexchange.config.ConfigTools;
-import com.openexchange.config.ConfigurationService;
+import com.openexchange.config.lean.LeanConfigurationService;
 import com.openexchange.configuration.ConfigurationExceptionCodes;
 import com.openexchange.exception.OXException;
 import com.openexchange.filestore.FileStorage;
 import com.openexchange.filestore.FileStorageProvider;
-import com.openexchange.filestore.utils.PropertyNameBuilder;
 import com.openexchange.java.Streams;
 import com.openexchange.java.Strings;
-import com.openexchange.server.ServiceExceptionCode;
 import com.openexchange.server.ServiceLookup;
 
 /**
@@ -146,25 +152,24 @@ public class S3FileStorageFactory implements FileStorageProvider {
     @Override
     public S3FileStorage getFileStorage(URI uri) throws OXException {
         try {
-            LOG.debug("Initializing S3 client for {}", uri);
-            /*
-             * extract filestore ID from authority part of URI
-             */
-            String filestoreID = extractFilestoreID(uri);
-            LOG.debug("Using \"{}\" as filestore ID.", filestoreID);
-            /*
-             * create client
-             */
-            ConfigurationService configService = services.getOptionalService(ConfigurationService.class);
-            if (null == configService) {
-                throw ServiceExceptionCode.absentService(ConfigurationService.class);
-            }
-            PropertyNameBuilder propNameBuilder = new PropertyNameBuilder("com.openexchange.filestore.s3.");
-            AmazonS3ClientInfo clientInfo = initClient(filestoreID, propNameBuilder, configService);
-            AmazonS3Client client = clientInfo.client;
-            String bucketName = initBucket(client, filestoreID, propNameBuilder, configService);
-            LOG.debug("Using \"{}\" as bucket name.", bucketName);
-            return new S3FileStorage(client, clientInfo.encrypted, bucketName, extractFilestorePrefix(uri), clientInfo.chunkSize);
+                LOG.debug("Initializing S3 client for {}", uri);
+                /*
+                 * extract filestore ID from authority part of URI
+                 */
+                String filestoreID = extractFilestoreID(uri);
+                LOG.debug("Using \"{}\" as filestore ID.", filestoreID);
+                /*
+                 * create client
+                 */
+                LeanConfigurationService configService = services.getServiceSafe(LeanConfigurationService.class);
+                Map<String, String> optional = getOptional(filestoreID);
+                String encryption = getPropertySafe(S3Properties.ENCRYPTION, configService, optional);
+                S3EncryptionConfig s3EncryptionConfig = new S3EncryptionConfig(encryption);
+                AmazonS3ClientInfo clientInfo = initClient(filestoreID, configService, s3EncryptionConfig);
+                AmazonS3Client client = clientInfo.client;
+                String bucketName = initBucket(client, filestoreID, configService, s3EncryptionConfig);
+                LOG.debug("Using \"{}\" as bucket name.", bucketName);
+                return new S3FileStorage(client, clientInfo.encrypted, s3EncryptionConfig.getServerSideEncryption() != null, bucketName, extractFilestorePrefix(uri), clientInfo.chunkSize);
         } catch (OXException ex) {
             Throwable cause = ex.getCause();
             if (cause instanceof AmazonS3Exception) {
@@ -198,25 +203,28 @@ public class S3FileStorageFactory implements FileStorageProvider {
      * Initializes an {@link AmazonS3Client} as configured by the referenced authority part of the supplied URI.
      *
      * @param filestoreID The filestore ID
+     * @param configService The {@link LeanConfigurationService}
+     * @param encryptionConfig The {@link S3EncryptionConfig} of the given filestore
      * @return The client
      * @throws OXException
      */
-    private AmazonS3ClientInfo initClient(String filestoreID, PropertyNameBuilder propNameBuilder, ConfigurationService configService) throws OXException {
+    private AmazonS3ClientInfo initClient(String filestoreID, LeanConfigurationService configService, S3EncryptionConfig encryptionConfig) throws OXException {
         /*
          * prepare credentials
          */
-        String accessKey = requireProperty(filestoreID, "accessKey", propNameBuilder, configService);
-        String secretKey = requireProperty(filestoreID, "secretKey", propNameBuilder, configService);
+        Map<String, String> optional = getOptional(filestoreID);
+        String accessKey = getPropertySafe(S3Properties.ACCESS_KEY, configService, optional);
+        String secretKey = getPropertySafe(S3Properties.SECRET_KEY, configService, optional);
+
         BasicAWSCredentials credentials = new BasicAWSCredentials(accessKey, secretKey);
         /*
          * instantiate client
          */
-        ClientConfiguration clientConfiguration = getClientConfiguration(filestoreID, propNameBuilder, configService);
+        ClientConfiguration clientConfiguration = getClientConfiguration(filestoreID, configService);
         AmazonS3Builder<?, ?> clientBuilder;
         boolean encrypted;
         {
-            String encryption = optProperty(filestoreID, "encryption", "none", propNameBuilder, configService);
-            if (Strings.isEmpty(encryption) || "none".equals(encryption)) {
+            if (encryptionConfig.getClientEncryption() == null || encryptionConfig.getClientEncryption().equals(EncryptionType.NONE)) {
                 AmazonS3ClientBuilder builder = AmazonS3ClientBuilder.standard()
                     .withCredentials(new AWSStaticCredentialsProvider(credentials))
                     .withClientConfiguration(clientConfiguration);
@@ -226,7 +234,7 @@ public class S3FileStorageFactory implements FileStorageProvider {
                 AmazonS3EncryptionClientBuilder builder = AmazonS3EncryptionClientBuilder.standard()
                     .withCredentials(new AWSStaticCredentialsProvider(credentials))
                     .withClientConfiguration(clientConfiguration)
-                    .withEncryptionMaterials(new StaticEncryptionMaterialsProvider(getEncryptionMaterials(filestoreID, encryption, propNameBuilder, configService)))
+                    .withEncryptionMaterials(new StaticEncryptionMaterialsProvider(getEncryptionMaterials(filestoreID, encryptionConfig.getClientEncryption(), configService)))
                     .withCryptoConfiguration(new CryptoConfiguration());
                 clientBuilder = builder;
                 encrypted = true;
@@ -235,23 +243,24 @@ public class S3FileStorageFactory implements FileStorageProvider {
         /*
          * configure client
          */
-        String endpoint = optProperty(filestoreID, "endpoint", null, propNameBuilder, configService);
-        if (false == Strings.isEmpty(endpoint)) {
+        String endpoint = configService.getProperty(S3Properties.ENDPOINT, optional);
+
+        if (Strings.isNotEmpty(endpoint)) {
             clientBuilder.setEndpointConfiguration(new EndpointConfiguration(endpoint, null));
         } else {
-            String region = optProperty(filestoreID, "region", "us-west-2", propNameBuilder, configService);
+            String region = configService.getProperty(S3Properties.REGION, optional);
             try {
                 clientBuilder.withRegion(Regions.fromName(region));
             } catch (IllegalArgumentException e) {
                 throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, region);
             }
         }
-        if (optBoolProperty(filestoreID, "pathStyleAccess", true, propNameBuilder, configService)) {
+        if (configService.getBooleanProperty(S3Properties.PATH_STYLE_ACCESS, optional)) {
             clientBuilder.setPathStyleAccessEnabled(Boolean.TRUE);
         }
         clientBuilder.withRequestHandlers(ETagCorrectionHandler.getInstance());
         long chunkSize;
-        String chunkSizeValue = optProperty(filestoreID, "chunkSize", "5MB", propNameBuilder, configService);
+        String chunkSizeValue = configService.getProperty(S3Properties.CHUNK_SIZE, optional);
         try {
             chunkSize = ConfigTools.parseBytes(chunkSizeValue);
         } catch (NumberFormatException e) {
@@ -260,10 +269,25 @@ public class S3FileStorageFactory implements FileStorageProvider {
         return new AmazonS3ClientInfo((AmazonS3Client) clientBuilder.build(), encrypted, chunkSize);
     }
 
-    private ClientConfiguration getClientConfiguration(String filestoreID, PropertyNameBuilder propNameBuilder, ConfigurationService configService) {
+    private String getPropertySafe(S3Properties prop, LeanConfigurationService configurationService, Map<String, String> optionals) throws OXException {
+        String property = configurationService.getProperty(prop, optionals);
+        if (Strings.isEmpty(property)) {
+            throw ConfigurationExceptionCodes.PROPERTY_MISSING.create(prop.getFQPropertyName(optionals));
+        }
+        return property;
+    }
+
+    private Map<String, String> getOptional(String filestoreId) {
+        HashMap<String, String> hashMap = new HashMap<>();
+        hashMap.put(S3Properties.OPTIONAL_NAME, filestoreId);
+        return hashMap;
+    }
+
+    private ClientConfiguration getClientConfiguration(String filestoreID, LeanConfigurationService configService) {
         ClientConfiguration clientConfiguration = new ClientConfiguration();
-        String signerOverride = optProperty(filestoreID, "signerOverride", "S3SignerType", propNameBuilder, configService);
-        if (false == Strings.isEmpty(signerOverride)) {
+        String signerOverride = configService.getProperty(S3Properties.SIGNER_OVERRIDE, getOptional(filestoreID));
+
+        if (Strings.isNotEmpty(signerOverride)) {
             clientConfiguration.setSignerOverride(signerOverride);
         }
 
@@ -291,13 +315,13 @@ public class S3FileStorageFactory implements FileStorageProvider {
         return clientConfiguration;
     }
 
-    private EncryptionMaterials getEncryptionMaterials(String filestoreID, String encryptionMode, PropertyNameBuilder propNameBuilder, ConfigurationService configService) throws OXException {
-        if (!"rsa".equalsIgnoreCase(encryptionMode)) {
-            throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Unknown encryption mode: " + encryptionMode);
+    private EncryptionMaterials getEncryptionMaterials(String filestoreID, EncryptionType clientType, LeanConfigurationService configurationService) throws OXException {
+        if (!EncryptionType.RSA.equals(clientType)) {
+            throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Unsupported encryption type: " + clientType.getName());
         }
-
-        String keyStore = requireProperty(filestoreID, "encryption.rsa.keyStore", propNameBuilder, configService);
-        String password = requireProperty(filestoreID, "encryption.rsa.password", propNameBuilder, configService);
+        Map<String, String> optional = getOptional(filestoreID);
+        String keyStore = getPropertySafe(S3Properties.RSA_KEYSTORE, configurationService, optional);
+        String password = getPropertySafe(S3Properties.RSA_PASSWORD, configurationService, optional);
         KeyPair keyPair = extractKeys(keyStore, password);
         return new EncryptionMaterials(keyPair);
     }
@@ -360,11 +384,14 @@ public class S3FileStorageFactory implements FileStorageProvider {
      *
      * @param s3client The S3 client
      * @param filestoreID The filestore ID
+     * @param configService The {@link LeanConfigurationService}
+     * @param encryptionConfig The {@link S3EncryptionConfig} for the given filestore
      * @return The bucket name
      * @throws OXException If initialization fails
      */
-    private String initBucket(AmazonS3Client s3client, String filestoreID, PropertyNameBuilder propNameBuilder, ConfigurationService configService) throws OXException {
-        String bucketName = requireProperty(filestoreID, "bucketName", propNameBuilder, configService);
+    private String initBucket(AmazonS3Client s3client, String filestoreID, LeanConfigurationService configService, S3EncryptionConfig encryptionConfig) throws OXException {
+        Map<String, String> optional = getOptional(filestoreID);
+        String bucketName = getPropertySafe(S3Properties.BUCKET_NAME, configService, optional);
 
         boolean bucketExists = false;
         try {
@@ -378,9 +405,14 @@ public class S3FileStorageFactory implements FileStorageProvider {
         }
 
         if (false == bucketExists) {
-            String region = optProperty(filestoreID, "region", "us-west-2", propNameBuilder, configService);
+            String region = configService.getProperty(S3Properties.REGION, getOptional(filestoreID));
+
             try {
                 s3client.createBucket(new CreateBucketRequest(bucketName, Region.fromValue(region)));
+
+                if (encryptionConfig.getServerSideEncryption() != null) {
+                    s3client.setBucketPolicy(new SetBucketPolicyRequest(bucketName, getSSEOnlyBucketPolicy(bucketName)));
+                }
             } catch (IllegalArgumentException e) {
                 throw S3ExceptionCode.UNEXPECTED_ERROR.create(e, e.getMessage());
             } catch (AmazonS3Exception e) {
@@ -401,6 +433,30 @@ public class S3FileStorageFactory implements FileStorageProvider {
         return bucketName;
     }
 
+    /**
+     * Gets the bucket policy for a server side encryption only bucket.
+     *
+     * @param bucket_name The name of the bucket
+     * @return The encryption only policy
+     */
+    private String getSSEOnlyBucketPolicy(String bucket_name) {
+        Policy bucket_policy = new Policy().withStatements(
+            new Statement(Statement.Effect.Deny)
+                .withId("DenyIncorrectEncryptionHeader")
+                .withPrincipals(Principal.AllUsers)
+                .withActions(S3Actions.PutObject)
+                .withResources(new Resource("arn:aws:s3:::" + bucket_name + "/*"))
+                .withConditions(new StringCondition(StringComparisonType.StringNotEquals, "s3:x-amz-server-side-encryption", "AES256")),
+            new Statement(Statement.Effect.Deny)
+                .withId("DenyUnEncryptedObjectUploads")
+                .withPrincipals(Principal.AllUsers)
+                .withActions(S3Actions.PutObject)
+                .withResources(new Resource("arn:aws:s3:::" + bucket_name + "/*"))
+                .withConditions(new BooleanCondition("s3:x-amz-server-side-encryption", true))
+                );
+        return bucket_policy.toJson();
+    }
+
     /**
      * Extracts the filestore prefix from the configured file store URI, i.e. the 'path' part of the URI.
      *
@@ -409,6 +465,7 @@ public class S3FileStorageFactory implements FileStorageProvider {
      * @throws IllegalArgumentException If the specified bucket name doesn't follow Amazon S3's guidelines
      */
     private static String extractFilestorePrefix(URI uri) throws IllegalArgumentException {
+        // Extract & prepare path to be used as prefix
         String path = uri.getPath();
         while (0 < path.length() && '/' == path.charAt(0)) {
             path = path.substring(1);
@@ -416,12 +473,9 @@ public class S3FileStorageFactory implements FileStorageProvider {
         if (path.endsWith("/")) {
             path = path.substring(0, path.length() - 1);
         }
-        /*
-         * Remove underscore characters to be conform to bucket name & prefix restrictions
-         * http://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html /
-         * http://docs.aws.amazon.com/AmazonS3/latest/dev/ListingKeysHierarchy.html
-         */
+
         if (path.endsWith("ctx_store")) {
+            // Expect context store identifier
             Matcher matcher = CTX_STORE_PATTERN.matcher(path);
             if (false == matcher.matches()) {
                 throw new IllegalArgumentException("Path does not match the expected pattern \"\\d+_ctx_store\" in URI: " + uri);
@@ -442,6 +496,14 @@ public class S3FileStorageFactory implements FileStorageProvider {
         return sanitizePathForPrefix(path, uri);
     }
 
+    /**
+     * Strips all characters from specified prefix path, which are no allows according to
+     * <a href="https://docs.aws.amazon.com/AmazonS3/latest/dev/UsingMetadata.html">this article</a>
+     *
+     * @param path The path to sanitize
+     * @param uri The file store URI containing the path
+     * @return The sanitized path ready to be used as prefix
+     */
     private static String sanitizePathForPrefix(String path, URI uri) {
         if (Strings.isEmpty(path)) {
             throw new IllegalArgumentException("Path is empty in URI: " + uri);
@@ -450,22 +512,40 @@ public class S3FileStorageFactory implements FileStorageProvider {
         StringBuilder sb = null;
         for (int k = path.length(), i = 0; k-- > 0; i++) {
             char ch = path.charAt(i);
-            if ('_' == ch) {
-                // Underscore not allowed
-                if (null == sb) {
-                    sb = new StringBuilder(path.length());
-                    sb.append(path, 0, i);
-                }
-            } else {
+            if (Strings.isAsciiLetterOrDigit(ch) || isAllowedSpecial(ch)) {
                 // Append
                 if (null != sb) {
                     sb.append(ch);
                 }
+            } else {
+                // Not allowed in prefix
+                if (null == sb) {
+                    sb = new StringBuilder(path.length());
+                    if (i > 0) {
+                        sb.append(path, 0, i);
+                    }
+                }
             }
         }
         return null == sb ? path : sb.toString();
     }
 
+    private static boolean isAllowedSpecial(char ch) {
+        switch (ch) {
+            case '!':
+            case '-':
+            case '_':
+            case '.':
+            case '*':
+            case '\'':
+            case '(':
+            case ')':
+                return true;
+            default:
+                return false;
+        }
+    }
+
     /**
      * Extracts the filestore ID from the configured file store URI, i.e. the 'authority' part from the URI.
      *
@@ -506,4 +586,89 @@ public class S3FileStorageFactory implements FileStorageProvider {
         }
     }
 
+    /**
+     *
+     * {@link S3EncryptionConfig} holds encryption types for client- and server-side encryption for a single filestore.
+     *
+     * @author <a href="mailto:kevin.ruthmann@open-xchange.com">Kevin Ruthmann</a>
+     * @since v7.10.1
+     */
+    private static class S3EncryptionConfig {
+
+        private final EncryptionType clientEncryption;
+        private final EncryptionType serverSideEncryption;
+
+        /**
+         * Initializes a new {@link S3FileStorageFactory.S3EncryptionConfig}.
+         *
+         * @throws OXException In case the configuration is invalid
+         */
+        S3EncryptionConfig(String config) throws OXException {
+            if (Strings.isEmpty(config)) {
+                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("An empty encryption type is invalid");
+            }
+            int index = config.indexOf('+');
+            if (index > 0) {
+                // E.g. rsa+sse-s3
+                String[] encryptionTypes = Strings.splitBy(config, '+', true);
+                if (encryptionTypes.length != 2) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("It's only allowed to combine one client side encryption type and one server side encryption type.");
+                }
+
+                EncryptionType typeA = EncryptionType.getTypeByName(encryptionTypes[0]);
+                if (typeA == null) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Unkown encryption type: " + encryptionTypes[0]);
+                }
+
+                EncryptionType typeB = EncryptionType.getTypeByName(encryptionTypes[1]);
+                if (typeB == null) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Unkown encryption type: " + encryptionTypes[1]);
+                }
+
+                if (typeA.isClientSideEncryption() == typeB.isClientSideEncryption()) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("It's only allowed to combine one client side encryption type and one server side encryption type.");
+                }
+
+                if (typeA.isClientSideEncryption()) {
+                    clientEncryption = typeA;
+                    serverSideEncryption = typeB;
+                } else {
+                    clientEncryption = typeB;
+                    serverSideEncryption = typeA;
+                }
+            } else {
+                EncryptionType type = EncryptionType.getTypeByName(config);
+                if (type == null) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Unkown encryption type: " + config);
+                }
+                if (type.isClientSideEncryption()) {
+                    clientEncryption = type;
+                    serverSideEncryption = null;
+                } else {
+                    clientEncryption = null;
+                    serverSideEncryption = type;
+                }
+            }
+        }
+
+        /**
+         * Gets the client encryption
+         *
+         * @return The client encryption
+         */
+        public EncryptionType getClientEncryption() {
+            return clientEncryption;
+        }
+
+        /**
+         * Gets the server-side encryption
+         *
+         * @return The server-side encryption
+         */
+        public EncryptionType getServerSideEncryption() {
+            return serverSideEncryption;
+        }
+
+    }
+
 }
